"""
 mod_axfr.py : Module, tries to get zone records

 Copyright (c) 2009, Karel Slany    (slany AT fit.vutbr.cz)
                     Zdenek Vasicek (vasicek AT fit.vutbr.cz)
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
 
     * Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
     * Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.
     * Neither the name of the organization nor the names of its
       contributors may be used to endorse or promote products derived from this
       software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.
"""

from interface import TestModule

import mod_get_dnservs
import mod_soa_recs

class Module(TestModule):
    """Tries to get zone records via AXFR

       The module uses data resolved by the mod_get_dnservs module. Module
       creates for each IP address entry a resolver. The resolvers are used to
       query the nameservers for zone transfers of their zone records.

       The module prints a summary of the obtained results. If you are not
       satisfied with the results. Try to increase the verbosity of the
       generated output.

       Be careful when creating IPv6 based resolvers. This may cause
       the application to crash on no IPv6 networks. Be also careful when
        trying to disable the IPv4 functionality.
    """

    def __init__(self, args, modenv):
        TestModule.__init__(self,args,modenv,"Get zone, AXFR")
        self.dns_env = self.get_module_modenv(mod_get_dnservs)
        self.soa_env = self.get_module_modenv(mod_soa_recs)

    def execute(self):

        res = {}
        
        # check option
        if (not self.args.options.zone):
            self.set_my_modenv(res)
            return True

        res["dnservs"] = {}
        domain = self.ldns.ldns_dname_new_frm_str(self.dns_env["domain"])
        resolver = self.ldns.ldns_resolver_new()

        # for each server name
        for servname in self.dns_env["dnservs"].keys():
            res["dnservs"][servname] = {}

            # ipv4
            if (not self.args.options.noipv4
                and ("ipv4" in self.dns_env["dnservs"][servname].keys())):
                res["dnservs"][servname]["ipv4"] = {}

                # for every ipv4 address
                for addr in self.dns_env["dnservs"][servname]["ipv4"].keys():
                    res["dnservs"][servname]["ipv4"][addr] = {}

                    if (not self.soa_env["dnservs"][servname]["ipv4"][addr]["alive"]):
                        continue

                    # create resolver
                    while resolver.pop_nameserver():
                        pass
                    resolver.push_nameserver(self.dns_env["dnservs"][servname]["ipv4"][addr]["rdf"])
                    if (not resolver):
                        continue

                    # ask for AXFR
                    status = resolver.axfr_start(domain, self.ldns.LDNS_RR_CLASS_IN)
                    res["dnservs"][servname]["ipv4"][addr]["AXFR"] = (status != self.ldns.LDNS_STATUS_OK)

            # ipv6
            if (self.args.options.resolver6
                and ("ipv6" in self.dns_env["dnservs"][servname].keys())):
                res["dnservs"][servname]["ipv6"] = {}

                # for every ipv4 address
                for addr in self.dns_env["dnservs"][servname]["ipv6"].keys():
                    res["dnservs"][servname]["ipv6"][addr] = {}

                    if (not self.soa_env["dnservs"][servname]["ipv6"][addr]["alive"]):
                        continue

                    # create resolver
                    while resolver.pop_nameserver():
                        pass
                    resolver.push_nameserver(self.dns_env["dnservs"][servname]["ipv6"][addr]["rdf"])
                    if (not resolver):
                        continue

                    # ask for AXFR
                    status = resolver.axfr_start(domain, self.ldns.LDNS_RR_CLASS_IN)
                    res["dnservs"][servname]["ipv6"][addr]["AXFR"] = (status != self.ldns.LDNS_STATUS_OK)


        # check whether all namervers for the domain don't allow AXFR
        all_restricted = True
        for servname in res["dnservs"].keys():
            if (not self.args.options.noipv4 and
                ("ipv4" in res["dnservs"][servname].keys())):
                for ipv4 in res["dnservs"][servname]["ipv4"].keys():
                    if ("AXFR" in res["dnservs"][servname]["ipv4"][ipv4].keys()):
                        restricted = not  res["dnservs"][servname]["ipv4"][ipv4]["AXFR"]
                        self.verb(1, servname, ipv4, "zone transfers:", not restricted)
                        all_restricted = all_restricted and restricted
            if (self.args.options.resolver6 and
                ("ipv6" in res["dnservs"][servname].keys())):
                for ipv6 in res["dnservs"][servname]["ipv6"].keys():
                    if ("AXFR" in res["dnservs"][servname]["ipv6"][ipv6].keys()):
                        restricted = not  res["dnservs"][servname]["ipv6"][ipv6]["AXFR"]
                        self.verb(1, servname, ipv6, "zone transfers:", not restricted)
                        all_restricted = all_restricted and restricted
        self.verb(1, "")
        if (all_restricted):
            self.verb(0, "All DNS servers have inactive AXFR transfers ( domain", self.dns_env["domain"], "): yes")
        else:
            self.verb(0, "All DNS servers have inactive AXFR transfers (domain", self.dns_env["domain"], "): no")
        self.verb(0, "")
        
        
        self.set_my_modenv(res)
        return True
