"""
 mod_dnssec_check.py : Module for resolving of A and AAAA records for given
 names

 Copyright (c) 2009, Karel Slany    (slany AT fit.vutbr.cz)
                     Zdenek Vasicek (vasicek AT fit.vutbr.cz)
                     Marek Vavrusa  (xvavru00 AT stud.fit.vutbr.cz)
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
 
     * Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
     * Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.
     * Neither the name of the organization nor the names of its
       contributors may be used to endorse or promote products derived from this
       software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.
"""

from interface import TestModule

DNSSEC_FAILURE  = 0
DNSSEC_BOGUS    = 1
DNSSEC_INSECURE = 2
DNSSEC_SECURE   = 3

DNSSEC_STR = [ "failure", "bogus", "insecure", "secure" ]

def dnsec_pkt_status(packet):
    if packet and packet.answer():
        # SERVFAIL indicated bogus name
        if (pkt.get_rcode() == ldns.LDNS_RCODE_SERVFAIL):
            return DNSSEC_BOGUS
        # Check AD (Authenticated) bit
        if (pkt.get_rcode() == ldns.LDNS_RCODE_NOERROR):
            if pkt.ad():
                return DNSSEC_SECURE
            else:
                return DNSSEC_INSECURE
        else:
            return DNSSEC_FAILURE
    else:
        return DNSSEC_FAILURE


class Module(TestModule):
    """Gets A records for all available nameserver addresses

       The module uses data resolved by the mod_get_dnservs module. Module
       creates for each DNS server entry (each DNS server IP address - a single
       server may have multiple addresses) a resolver. Then asks by using the
       resolvers for A and/or AAAA entries - for a specific name to be
       translated.

       The obtained A and AAAA records are stored in a tree structure created
       from python dictionaries. This structure is copied into a global
       environment variable, on order to keep the results accessible for
       other modules.

       The module tests whether all the available nameservers contain the
       requested A and/or AAAA record. If a n record is found, the module
       tests whether the answer was received as an authoritative answer.
       If you are not satisfied with the outputs, you can always try to
       increase the verbosity of the module outputs.

       Be careful when creating IPv6 based resolvers. This may cause
       the application to crash on no IPv6 networks. Be also careful when
       trying to disable the IPv4 functionality.
    """

    def __init__(self, args, modenv):
        TestModule.__init__(self,args,modenv,"Check DNSSEC")

    def execute(self):

        res = {}

        # check option
        if ((not self.args.options.arecs)
            or (not self.args.options.vresolver)):
            self.set_my_modenv(res)
            return True

        # create a validating resolver
        vresolver = self.ldns.ldns_resolver.new_frm_file(None)
        vresolver.set_dnssec(True)
        self.verb(2, "using validating resolver")
        while vresolver.pop_nameserver():
            pass
        for ipstr in self.args.options.vresolver:
            if (ipstr.find(":") == -1):
                # ipv4
                self.verb(2, "adding to resolver", ipstr)
                ip = self.ldns.ldns_rdf.new_frm_str(ipstr,
                    self.ldns.LDNS_RDF_TYPE_A)
            else:
                # ipv6
                self.verb(2, "adding to resolver", ipstr)
                ip = self.ldns.ldns_rdf.new_frm_str(ipstr,
                    self.ldns.LDNS_RDF_TYPE_AAAA)
            vresolver.push_nameserver(ip)

        self.verb(2, "A and AAAA records:")

        res["dnssec"] = {}

        for aname in self.args.options.arecs:
            res["dnssec"][aname] = {}
            res["dnssec"][aname]["ipv4"] = {}
            res["dnssec"][aname]["ipv4"]["dnssec-secure"] = False
            res["dnssec"][aname]["ipv4"]["A"] = []
            res["dnssec"][aname]["ipv6"] = {}
            res["dnssec"][aname]["ipv6"]["dnssec-secure"] = False
            res["dnssec"][aname]["ipv6"]["AAAA"] = []

            if (not self.args.options.noipv4):
                pkt = vresolver.query(aname, self.ldns.LDNS_RR_TYPE_A,
                    self.ldns.LDNS_RR_CLASS_IN)
                res["dnssec"][aname]["ipv4"]["dnssec-secure"] = dnsec_pkt_status == DNSSEC_SECURE
                if (pkt):
                    arec = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_A,
                        self.ldns.LDNS_SECTION_ANSWER)
                    if (arec):
                        for arec_rr in arec.rrs():
                            self.verb(2, aname, arec_rr.rdf(0),
                                "DNSSEC secure:", res["dnssec"][aname]["ipv4"]["dnssec-secure"])
                            res["dnssec"][aname]["ipv4"]["A"].append(str(arec_rr.rdf(0)))
            if (self.args.options.ipv6):
                pkt = vresolver.query(aname, self.ldns.LDNS_RR_TYPE_AAAA,
                    self.ldns.LDNS_RR_CLASS_IN)
                res["dnssec"][aname]["ipv6"]["dnssec-secure"] = dnsec_pkt_status == DNSSEC_SECURE
                arec = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_AAAA,
                    self.ldns.LDNS_SECTION_ANSWER)
                if (arec):
                    for arec_rr in arec.rrs():
                        self.verb(2, aname, arec_rr.rdf(0),
                            "DNSSEC secure:", res["dnssec"][aname]["ipv6"]["dnssec-secure"])
                        res["dnssec"][aname]["ipv6"]["AAAA"].append(str(arec_rr.rdf(0)))
        self.verb(2, "")

        # summarize results
        all_secure = True
        for aname in res["dnssec"].keys():
            if (not self.args.options.noipv4):
                if (res["dnssec"][aname]["ipv4"]["dnssec-secure"]):
                    self.verb(1, "for", aname, "received DNSSEC secured ipv4:", *res["dnssec"][aname]["ipv4"]["A"])
                else:
                    self.verb(1, "for", aname, "received DNSSEC not secured ipv4:", *res["dnssec"][aname]["ipv4"]["A"])
                all_secure = all_secure and res["dnssec"][aname]["ipv4"]["dnssec-secure"]
            if (self.args.options.ipv6):
                if (res["dnssec"][aname]["ipv6"]["dnssec-secure"]):
                    self.verb(1, "for", aname, "received DNSSEC secured ipv6:", *res["dnssec"][aname]["ipv6"]["AAAA"])
                else:
                    self.verb(1, "for", aname, "received DNSSEC not secured ipv6:", *res["dnssec"][aname]["ipv6"]["AAAA"])
                all_secure = all_secure and res["dnssec"][aname]["ipv6"]["dnssec-secure"]
        self.verb(1, "")
        if (all_secure):
            self.verb(0, "All answers DNSSEC secured by given resolver: yes")
        else:
            self.verb(0, "All answers DNSSEC secured by given resolver: no")
        self.verb(0, "")

        self.set_my_modenv(res)
        return True
