"""
 mod_mx_recs.py : Module, checks for MX entries

 Copyright (c) 2009, Karel Slany    (slany AT fit.vutbr.cz)
                     Zdenek Vasicek (vasicek AT fit.vutbr.cz)
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
 
     * Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
     * Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.
     * Neither the name of the organization nor the names of its
       contributors may be used to endorse or promote products derived from this
       software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.
"""

from interface import TestModule

import mod_get_dnservs
import mod_soa_recs

class Module(TestModule):
    """Gets MX records

       The module uses data resolved by the mod_get_dnservs module. Module
       creates for each DNS server IP address a resolver. Then asks by using
       the resolvers for MX entries - for a specific name to be
       translated.

       A simple test is performed. It compares all records and prints out
       a warning if some records do not match.
 
       The module tests whether all the available nameservers contain the
       Be careful when creating IPv6 based resolvers. This may cause
       the application to crash on no IPv6 networks. Be also careful when
       trying to disable the IPv4 functionality.
    """

    def __init__(self, args, modenv):
        TestModule.__init__(self,args,modenv,"Get MX records")
        self.dns_env = self.get_module_modenv(mod_get_dnservs)
        self.soa_env = self.get_module_modenv(mod_soa_recs)

    def execute(self):

        res = {}        

        # check option
        if (not self.args.options.mxrecs):
            self.set_my_modenv(res)
            return True

        res["dnservs"] = {}
        domain = self.ldns.ldns_dname_new_frm_str(self.dns_env["domain"])
        resolver = self.ldns.ldns_resolver_new()

        # for each server name
        for servname in self.dns_env["dnservs"].keys():
            res["dnservs"][servname] = {}

            # for ipv4 addresses
            if (not self.args.options.noipv4
                and ("ipv4" in self.dns_env["dnservs"][servname].keys())):
                res["dnservs"][servname]["ipv4"] = {}

                # for every address
                for addr in self.dns_env["dnservs"][servname]["ipv4"].keys():
                    res["dnservs"][servname]["ipv4"][addr] = {}

                    if (not self.soa_env["dnservs"][servname]["ipv4"][addr]["alive"]):
                        continue

                    # create resolver
                    while resolver.pop_nameserver():
                        pass
                    resolver.push_nameserver(self.dns_env["dnservs"][servname]["ipv4"][addr]["rdf"])
                    if (not resolver):
                        continue

                    # ask for MX
                    pkt = resolver.query(domain, self.ldns.LDNS_RR_TYPE_MX,
                        self.ldns.LDNS_RR_CLASS_IN,self.ldns.LDNS_RD)
                    if (not pkt):
                        self.err(0, "Resolver response error")
                        return False
                    mx = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_MX,
                        self.ldns.LDNS_SECTION_ANSWER)
                    if (not self.args.options.noipv4):
                        mxadi4 = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_A,
                            self.ldns.LDNS_SECTION_ADDITIONAL)
                    if (self.args.options.ipv6):
                        mxadi6 = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_AAAA,
                            self.ldns.LDNS_SECTION_ADDITIONAL)
                    self.verb(2, "MX records for", servname, "ipv4", addr)
                    if (mx):
                        # parse results
                        mx.sort()
                        res["dnservs"][servname]["ipv4"][addr]["MX"] = []
                        for mx_rr in mx.rrs():
                            record = {}
                            record["owner"] = str(mx_rr.owner())
                            record["ttl"] = int(mx_rr.ttl())
                            record["server-pri"] = int(str(mx_rr.rdf(0)))
                            record["server-name"] = str(mx_rr.rdf(1))
                            record["ipv4"] = None
                            record["ipv6"] = None
                            if ((not self.args.options.noipv4) and mxadi4):
                                for mxadi4_rr in mxadi4.rrs():
                                    if (mxadi4_rr.owner() == mx_rr.rdf(1)):
                                        record["ipv4"] = str(mxadi4_rr.a_address())
                            if ((self.args.options.ipv6) and mxadi6):
                                for mxadi6_rr in mxadi6.rrs():
                                    if (mxadi6_rr.owner() == mx_rr.rdf(1)):
                                        record["ipv6"] = str(mxadi6_rr.a_address())
                            res["dnservs"][servname]["ipv4"][addr]["MX"].append(record)
                            if (record["ipv4"] and record["ipv6"]):
                                self.verb(2, "owner:", record["owner"],
                                    "ttl:", record["ttl"],
                                    "server:", record["server-pri"], record["server-name"],
                                    "ipv4:", record["ipv4"],
                                    "ipv6:", record["ipv6"])
                            elif (record["ipv4"]):
                                self.verb(2, "owner:", record["owner"],
                                    "ttl:", record["ttl"],
                                    "server:", record["server-pri"], record["server-name"],
                                    "ipv4:", record["ipv4"])
                            elif (record["ipv6"]):
                                self.verb(2, "owner:", record["owner"],
                                    "ttl:", record["ttl"],
                                    "server:", record["server-pri"], record["server-name"],
                                    "ipv6:", record["ipv6"])
                            else:
                                self.verb(2, "owner:", record["owner"],
                                    "ttl:", record["ttl"],
                                    "server:", record["server-pri"], record["server-name"])
                    else:
                        self.verb(2, "none")
                    self.verb(2, "")

            # for ipv6 addresses
            if (self.args.options.resolver6
                and ("ipv6" in self.dns_env["dnservs"][servname].keys())):
                res["dnservs"][servname]["ipv6"] = {}
                
                # for every address
                for addr in self.dns_env["dnservs"][servname]["ipv6"].keys():
                    res["dnservs"][servname]["ipv6"][addr] = {}
 
                    if (not self.soa_env["dnservs"][servname]["ipv6"][addr]["alive"]):
                        continue

                    # create resolver
                    while resolver.pop_nameserver():
                        pass
                    resolver.push_nameserver(self.dns_env["dnservs"][servname]["ipv6"][addr]["rdf"])
                    if (not resolver):
                        continue

                    # ask for MX
                    pkt = resolver.query(domain, self.ldns.LDNS_RR_TYPE_MX,
                        self.ldns.LDNS_RR_CLASS_IN,self.ldns.LDNS_RD)
                    if (not pkt):
                        self.err(0, "Resolver response error")
                        return False
                    mx = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_MX,
                        self.ldns.LDNS_SECTION_ANSWER)
                    if (not self.args.options.noipv4):
                        mxadi4 = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_A,
                            self.ldns.LDNS_SECTION_ADDITIONAL)
                    if (self.args.options.ipv6):
                        mxadi6 = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_AAAA,
                            self.ldns.LDNS_SECTION_ADDITIONAL)
                    self.verb(2, "MX records for", servname, "ipv6", addr)
                    if (mx):
                        # parse results
                        mx.sort()
                        res["dnservs"][servname]["ipv6"][addr]["MX"] = []
                        for mx_rr in mx.rrs():
                            record = {}
                            record["owner"] = str(mx_rr.owner())
                            record["ttl"] = int(mx_rr.ttl())
                            record["server-pri"] = int(str(mx_rr.rdf(0)))
                            record["server-name"] = str(mx_rr.rdf(1))
                            record["ipv4"] = None
                            record["ipv6"] = None
                            if ((not self.args.options.noipv4) and mxadi4):
                                for mxadi4_rr in mxadi4.rrs():
                                    if (mxadi4_rr.owner() == mx_rr.rdf(1)):
                                        record["ipv4"] = str(mxadi4_rr.a_address())
                            if ((self.args.options.ipv6) and mxadi6):
                                for mxadi6_rr in mxadi6.rrs():
                                    if (mxadi6_rr.owner() == mx_rr.rdf(1)):
                                        record["ipv4"] = str(mxadi6_rr.a_address())
                            res["dnservs"][servname]["ipv6"][addr]["MX"].append(record)
                            if (record["ipv4"] and record["ipv6"]):
                                self.verb(2, "owner:", record["owner"],
                                    "ttl:", record["ttl"],
                                    "server:", record["server-pri"], record["server-name"],
                                    "ipv4:", record["ipv4"],
                                    "ipv6:", record["ipv6"])
                            elif (record["ipv4"]):
                                self.verb(2, "owner:", record["owner"],
                                    "ttl:", record["ttl"],
                                    "server:", record["server-pri"], record["server-name"],
                                    "ipv4:", record["ipv4"])
                            elif (record["ipv6"]):
                                self.verb(2, "owner:", record["owner"],
                                    "ttl:", record["ttl"],
                                    "server:", record["server-pri"], record["server-name"],
                                    "ipv6:", record["ipv6"])
                            else:
                                self.verb(2, "owner:", record["owner"],
                                    "ttl:", record["ttl"],
                                    "server:", record["server-pri"], record["server-name"])
                    else:
                        self.verb(2, "none")
                    self.verb(2, "")

        # check if all records are same
        owner = {}
        ttl = {}
        pri = {}
        some_empty = False
        no_record = True
        added = False
        missing = False
        for servname in res["dnservs"].keys():
            if ((not self.args.options.noipv4) and
                ("ipv4" in res["dnservs"][servname].keys())):
                for ipv4 in res["dnservs"][servname]["ipv4"].keys():
                    if ("MX" in res["dnservs"][servname]["ipv4"][ipv4].keys()):
                        if (no_record):
                            # create first entry
                            no_record = False
                            for mxrec in res["dnservs"][servname]["ipv4"][ipv4]["MX"]:
                                name = mxrec["server-name"]
                                owner[name] = []
                                owner[name].append(mxrec["owner"])
                                ttl[name] = []
                                ttl[name].append(mxrec["ttl"])
                                pri[name] = []
                                pri[name].append(mxrec["server-pri"])
                                self.verb(1, servname, ipv4, ":", mxrec["server-pri"], name)
                        else:
                            # compare with existing entries
                            for mxrec in res["dnservs"][servname]["ipv4"][ipv4]["MX"]:
                                name = mxrec["server-name"]
                                if (name in owner.keys()):
                                    if (mxrec["owner"] not in owner[name]):
                                        owner[name].append(mxrec["owner"])
                                    if (mxrec["ttl"] not in ttl[name]):
                                        ttl[name].append(mxrec["ttl"])
                                    if (mxrec["server-pri"] not in pri[name]):
                                        pri[name].append(mxrec["server-pri"])
                                else:
                                    added = True
                                    owner[name] = []
                                    owner[name].append(mxrec["owner"])
                                    ttl[name] = []
                                    ttl[name].append(mxrec["ttl"])
                                    pri[name] = []
                                    pri[name].append(mxrec["server-pri"])
                                self.verb(1, servname, ipv4, ":", mxrec["server-pri"], name)
                            if (len(owner.keys()) !=
                                len(res["dnservs"][servname]["ipv4"][ipv4]["MX"])):
                                mising = True
                    else:
                        self.verb(1, servname, ipv4, "No MX records")
                        some_empty = True

            if (self.args.options.resolver6 and
                ("ipv6" in res["dnservs"][servname].keys())):
                for ipv6 in res["dnservs"][servname]["ipv6"].keys():
                    if ("MX" in res["dnservs"][servname]["ipv6"][ipv6].keys()):
                        if (no_record):
                            # create first entry
                            no_record = False
                            for mxrec in res["dnservs"][servname]["ipv6"][ipv6]["MX"]:
                                name = mxrec["server-name"]
                                owner[name] = []
                                owner[name].append(mxrec["owner"])
                                ttl[name] = []
                                ttl[name].append(mxrec["ttl"])
                                pri[name] = []
                                pri[name].append(mxrec["server-pri"])
                                self.verb(1, servname, ipv6, ":", mxrec["server-pri"], name)
                        else:
                            # compare with existing entries
                            for mxrec in res["dnservs"][servname]["ipv6"][ipv6]["MX"]:
                                name = mxrec["server-name"]
                                if (name in owner.keys()):
                                    if (mxrec["owner"] not in owner[name]):
                                        owner[name].append(mxrec["owner"])
                                    if (mxrec["ttl"] not in ttl[name]):
                                        ttl[name].append(mxrec["ttl"])
                                    if (mxrec["server-pri"] not in pri[name]):
                                        pri[name].append(mxrec["server-pri"])
                                else:
                                    added = True
                                    owner[name] = []
                                    owner[name].append(mxrec["owner"])
                                    ttl[name] = []
                                    ttl[name].append(mxrec["ttl"])
                                    pri[name] = []
                                    pri[name].append(mxrec["server-pri"])
                                self.verb(1, servname, ipv6, ":", mxrec["server-pri"], name)
                            if (len(owner.keys()) !=
                                len(res["dnservs"][servname]["ipv6"][ipv6]["MX"])):
                                mising = True
                    else:
                        self.verb(1, servname, ipv6, "No MX records")
                        some_empty = True
        self.verb(1, "")
        # process data obtained from check
        owner_same = True
        ttl_same = True
        pri_same = True
        for name in owner.keys():
            owner_same = owner_same and (len(owner[name]) == 1)
            ttl_same = ttl_same and (len(ttl[name]) == 1)
            pri_same = pri_same and (len(pri[name]) == 1)
        if (no_record):
            self.verb(0, "MX records found: no")
        elif (some_empty or added or missing):
            self.verb(0, "MX records found: yes")
            self.verb(0, "")
            self.verb(0, "Equal nameservers listed: no")
        else:
            self.verb(0, "MX records found: yes")
            self.verb(0, "")
            self.verb(0, "Equal nameservers listed: yes")
            self.verb(0, "")
            if (owner_same and ttl_same and pri_same):
                self.verb(0, "Equal MX records on nameservers: yes")
            else:
                self.verb(0, "Equal MX records on nameservers: no")
        self.verb(0, "")

        self.set_my_modenv(res)
        return True
