"""
 mod_ptr_recs.py : Module tries to reverse resolve domain names
 from name server ips. Then performs a check whether the obtained
 names are resolvable back to the original ips.

 Copyright (c) 2009, Karel Slany    (slany AT fit.vutbr.cz)
                     Zdenek Vasicek (vasicek AT fit.vutbr.cz)
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
 
     * Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
     * Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.
     * Neither the name of the organization nor the names of its
       contributors may be used to endorse or promote products derived from this
       software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.
"""

from interface import TestModule

import mod_get_dnservs

def reverse4(str4):
    # the reverted ip ends with "."
    return ".".join(str4.split(".")[::-1])+"."

def reverse6(str6):
    # the reverted ip ends with "."
    b = ""
    for a in str6.split(":"):
        if (not a):
            a = "0"
        a = int(a, 16)
        b = ".".join(("%04x" % a)[::-1]) + "." + b
    return b

class Module(TestModule):
    """Resolve PTR for ip adresses

       The module uses data resolved by the mod_get_dnservs module. A resolver
       using system default settings is created. This resolver is used for
       reverse resolution of domain names from all IPs of the nameserves, which
       are obtained prom the mod_get_dnservs module.

       The reverse resolved names are again translated onto IPs.

       The test in this module tries to resolve a reverse name from an IP
       and then back resolve it to an IP address. The original and resolved IPs
       are compared. If a mismatch appears, a message will be printed.

       Currently if no domain names from IPs could not be resolved the module
       will not generate any output.

       In cases when a domain name could not be back resolved to IP, the module
       will generate a message.

       Currently only IPv4 functionality is implemented.

       If you are not satisfied with the generated outputs, try increasing
       verbosity.
    """

    def __init__(self, args, modenv):
        TestModule.__init__(self,args,modenv,"Get PTR records")
        self.dns_env = self.get_module_modenv(mod_get_dnservs)

    def execute(self):

        res = {}

        res["dnservs"] = {}
        stat, resolver = self.ldns.ldns_resolver_new_frm_file(None)
        if (not resolver):
            raise Exception("Can't create resolver. Error:%s" % stat)

        # each nameserver
        for servname in self.dns_env["dnservs"].keys():
            res["dnservs"][servname] = {}

            # for ipv4 addresses
            if (not self.args.options.noipv4
                and ("ipv4" in self.dns_env["dnservs"][servname].keys())):
                res["dnservs"][servname]["ipv4"] = {}

                # for every address
                for addr in self.dns_env["dnservs"][servname]["ipv4"].keys():
                    res["dnservs"][servname]["ipv4"][addr] = {}

                    # ask resolver for reverse records
                    query = reverse4(addr) + "in-addr.arpa."
                    pkt = resolver.query(query,
                        self.ldns.LDNS_RR_TYPE_PTR,
                        self.ldns.LDNS_RR_CLASS_IN,
                        self.ldns.LDNS_RD)
                    if (not pkt):
                        self.err(0, "Resolver response error")
                        return False
                    ptrrec = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_PTR,
                        self.ldns.LDNS_SECTION_ANSWER)
                    if (not ptrrec):
                        continue
                    # for every PTR record in answer
                    for ptrrec_rr in ptrrec.rrs():
                        if ("PTR" not in res["dnservs"][servname]["ipv4"][addr].keys()):
                            res["dnservs"][servname]["ipv4"][addr]["PTR"] = {}
                        name = str(ptrrec_rr.rdf(0))
                        # write down name and ask for ips
                        res["dnservs"][servname]["ipv4"][addr]["PTR"][name] = {}
                        res["dnservs"][servname]["ipv4"][addr]["PTR"][name]["aa"] = pkt.aa()
                        # resolve ipv4 adresses
                        if (not self.args.options.noipv4):
                            res["dnservs"][servname]["ipv4"][addr]["PTR"][name]["ipv4"] = {}
                            res["dnservs"][servname]["ipv4"][addr]["PTR"][name]["ipv4"]["failed"] = False
                            res["dnservs"][servname]["ipv4"][addr]["PTR"][name]["ipv4"]["addr"] = []
                            res["dnservs"][servname]["ipv4"][addr]["PTR"][name]["ipv4"]["qlist"] = []
                            res["dnservs"][servname]["ipv4"][addr]["PTR"][name]["ipv4"]["qlist"].append(name)
                            subpkt = resolver.query(name,
                                self.ldns.LDNS_RR_TYPE_A,
                                self.ldns.LDNS_RR_CLASS_IN,
                                self.ldns.LDNS_RD)
                            subaddr = subpkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_A,
                                self.ldns.LDNS_SECTION_ANSWER)
                            loop_counter = 0
                            while ((not subaddr) and (loop_counter < 10)):
                                # answer did not come, istead test CNAME and try to redirect
                                loop_counter += 1
                                cname = subpkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_CNAME,
                                    self.ldns.LDNS_SECTION_ANSWER)
                                if (not cname):
                                    break
                                for cname_rr in cname.rrs():
                                    if (str(cname_rr.rdf(0)) in res["dnservs"][servname]["ipv4"][addr]["PTR"][name]["ipv4"]["qlist"]):
                                        continue
                                    self.verb(2, "redirecting via cname", servname, addr, name, "->", cname_rr.rdf(0))
                                    res["dnservs"][servname]["ipv4"][addr]["PTR"][name]["ipv4"]["qlist"].append(str(cname_rr.rdf(0)))
                                    subpkt = resolver.query(cname_rr.rdf(0),
                                        self.ldns.LDNS_RR_TYPE_A,
                                        self.ldns.LDNS_RR_CLASS_IN,
                                        self.ldns.LDNS_RD)
                                    subaddr = subpkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_A,
                                        self.ldns.LDNS_SECTION_ANSWER)
                                    if (subaddr):
                                        break
                            if (subaddr):
                                for subaddr_rr in subaddr.rrs():
                                    self.verb(2, servname, addr, "reverse:", name, subaddr_rr.rdf(0))
                                    res["dnservs"][servname]["ipv4"][addr]["PTR"][name]["ipv4"]["addr"].append(str(subaddr_rr.rdf(0)))
                            else:
                                self.verb(2, servname, addr, "failed to get reverse name and its ipv4 via",
                                    *(res["dnservs"][servname]["ipv4"][addr]["PTR"][name]["ipv4"]["qlist"]))
                                # lookup failed
                                res["dnservs"][servname]["ipv4"][addr]["PTR"][name]["ipv4"]["failed"] = True
                        # resolves ipv6 ways
                        if (self.args.options.ipv6):
                            res["dnservs"][servname]["ipv4"][addr]["PTR"][name]["ipv6"] = {}
                            res["dnservs"][servname]["ipv4"][addr]["PTR"][name]["ipv6"]["failed"] = False
                            res["dnservs"][servname]["ipv4"][addr]["PTR"][name]["ipv6"]["addr"] = []
                            res["dnservs"][servname]["ipv4"][addr]["PTR"][name]["ipv6"]["qlist"] = []
                            res["dnservs"][servname]["ipv4"][addr]["PTR"][name]["ipv6"]["qlist"].append(name)
                            

            # for ipv6 addresses
            if (self.args.options.resolver6
                and ("ipv6" in self.dns_env["dnservs"][servname].keys())):
                res["dnservs"][servname]["ipv6"] = {}



        self.verb(2, "")

        # check whether ip adresses are reverse resolvable
        # (if exists a reverse record and if one exist,
        # compares its addres with the original)
        b_resolvable = {}
        b_resolvable["dnservs"] = {}
        for servname in res["dnservs"].keys():
            b_resolvable["dnservs"][servname] = {}
            if ((not self.args.options.noipv4) and
                ("ipv4" in res["dnservs"][servname].keys())):
                b_resolvable["dnservs"][servname]["ipv4"] = {}
                # all paths resolved
                b_resolvable["dnservs"][servname]["ipv4"]["all"] = len(res["dnservs"][servname]["ipv4"].keys()) != 0
                # all pths mathc in result
                b_resolvable["dnservs"][servname]["ipv4"]["all_match"] = len(res["dnservs"][servname]["ipv4"].keys()) != 0
                # only some resolved
                b_resolvable["dnservs"][servname]["ipv4"]["some"] = False
                # only some match
                b_resolvable["dnservs"][servname]["ipv4"]["some_match"] = False
                for ipv4 in res["dnservs"][servname]["ipv4"].keys():
                    if ("PTR" in res["dnservs"][servname]["ipv4"][ipv4].keys()):
                        # test every reverse name entry
                        for name in res["dnservs"][servname]["ipv4"][ipv4]["PTR"].keys():
                            failed = res["dnservs"][servname]["ipv4"][ipv4]["PTR"][name]["ipv4"]["failed"]
                            found_original = ipv4 in res["dnservs"][servname]["ipv4"][ipv4]["PTR"][name]["ipv4"]["addr"]
                            if (failed):
                                self.verb(1, servname, "ipv4", ipv4, "leads to failed resolving via (in order):",
                                    *res["dnservs"][servname]["ipv4"][ipv4]["PTR"][name]["ipv4"]["qlist"])
                            elif (not found_original):
                                self.verb(1, servname, "ipv4", ipv4, "found ip (not original):",
                                    *res["dnservs"][servname]["ipv4"][ipv4]["PTR"][name]["ipv4"]["addr"])
                            else:
                                self.verb(1, servname, "ipv4", ipv4, "found ip:",
                                    *res["dnservs"][servname]["ipv4"][ipv4]["PTR"][name]["ipv4"]["addr"])
                            # all reverse names lead to an ip
                            b_resolvable["dnservs"][servname]["ipv4"]["all"] = (
                                b_resolvable["dnservs"][servname]["ipv4"]["all"]
                                and not failed)
                            # all reverse names lead to original ip
                            b_resolvable["dnservs"][servname]["ipv4"]["all_match"] = (
                                b_resolvable["dnservs"][servname]["ipv4"]["all_match"]
                                and found_original)
                            # some reverse names lead to an ip
                            b_resolvable["dnservs"][servname]["ipv4"]["some"] = (
                                b_resolvable["dnservs"][servname]["ipv4"]["some"]
                                or not failed)
                            # some reverse names lead to original ip
                            b_resolvable["dnservs"][servname]["ipv4"]["some_match"] = (
                                b_resolvable["dnservs"][servname]["ipv4"]["some_match"]
                                or found_original)
                    else:
                        b_resolvable["dnservs"][servname]["ipv4"]["all"] = False
                        b_resolvable["dnservs"][servname]["ipv4"]["some"] = False
            if (self.args.options.resolver6 and
                ("ipv6" in res["dnservs"][servname].keys())):
                b_resolvable["dnservs"][servname]["ipv6"] = {}
                b_resolvable["dnservs"][servname]["ipv6"]["all"] = len(res["dnservs"][servname]["ipv6"].keys()) != 0
                b_resolvable["dnservs"][servname]["ipv6"]["some"] = False
                for ipv4 in res["dnservs"][servname]["ipv6"].keys():
                    pass

        self.verb(1, "")

        # test whether all ways are or only some aresolvable
        b_resolvable_all_match = True
        b_resolvable_some_match = True
        for servname in b_resolvable["dnservs"].keys():
            if ((not self.args.options.noipv4) and
                ("ipv4" in res["dnservs"][servname].keys())):
                b_resolvable_all_match = b_resolvable_all_match and b_resolvable["dnservs"][servname]["ipv4"]["all_match"]
                b_resolvable_some_match = b_resolvable_some_match and b_resolvable["dnservs"][servname]["ipv4"]["some_match"]
            if (self.args.options.resolver6 and
                ("ipv6" in res["dnservs"][servname].keys())):
                pass
        if (b_resolvable_all_match):
            self.verb(0, "Each reverse-resolved name for each ipv4 is resolvable back to its original ip: yes")
        else:
            self.verb(0, "Each reverse-resolved name for each ipv4 is resolvable back to its original ip: no")
        self.verb(0, "")
        if (b_resolvable_some_match):
            self.verb(0, "At least one reverse-resolved name for each ipv4 is resolvable back to its original ip: yes")
        else:
            self.verb(0, "At least one reverse-resolved name for each ipv4 is resolvable back to its original ip: no")
        self.verb(0, "")
        
        self.set_my_modenv(res)
        return True
