"""
 mod_soa_recs.py : Module checks the SOA records on the nameservers
 for common errors

 Copyright (c) 2009, Karel Slany    (slany AT fit.vutbr.cz)
                     Zdenek Vasicek (vasicek AT fit.vutbr.cz)
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
 
     * Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
     * Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.
     * Neither the name of the organization nor the names of its
       contributors may be used to endorse or promote products derived from this
       software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.
"""

from interface import TestModule

import mod_get_dnservs

def utf2name(name):
    splited = name.split('.')
    while (u"" in splited):
        splited.remove(u"")
    restored = '.'.join([encodings.idna.ToASCII(a) for a in splited])
    return restored + "."

def name_syntax(name, dname):
    # check whether the actual domain is not repeated twice
    name = name.replace(dname, "", 1)
    return not name.endswith(dname)

soa_keys = ["name-server","email-addr","sn","refresh","retry","expiry","min"]

class Module(TestModule):
    """Gets SOA records

       The module uses data resolved by the mod_get_dnservs module. Module
       creates for each DNS server IP a resolver. Then asks by using the
       resolvers for SOA entries.

       The obtained SOA records are stored in a tree structure created
       from python dictionaries. This structure is copied into a global
       environment variable, on order to keep the results accessible for
       other modules.

       The module preforms various tests, checking the SOA records for some
       common mistakes. The values are compared between themselves and in case
       of a contradiction a message is printed.

       Currently in checks the serial number, whether the nameservers are responding with
       a SOA, or are alive. It checks the mname entry in SOA, for a domain duplicity (eg.
       tested domain is nic.cz. and the primary server name ends with nic.cz.nic.cz.)
       This may be caused by a wrong server configuration. Itl aseo checks the email entry
       in SOA. It skould not contain a '@'.

       Be careful when creating IPv6 based resolvers. This may cause
       the application to crash on no IPv6 networks. Be also careful when
       trying to disable the IPv4 functionality.

    """

    def __init__(self, args, modenv):
        TestModule.__init__(self,args,modenv,"Get SOA records")
        # depends on mod_get_dnservs
        self.dns_env = self.get_module_modenv(mod_get_dnservs)

    def execute(self):
        # first, create tree structure holding SOA records of all available servers

        # for each ip adress create a resolver
        res = {}
        res["dnservs"] = {}
        domain = self.ldns.ldns_dname_new_frm_str(self.dns_env["domain"])
        resolver = self.ldns.ldns_resolver_new()

        self.verb(2, "SOA records")

        # for each server name
        for servname in self.dns_env["dnservs"].keys():
            res["dnservs"][servname] = {}

            # for ipv4 addresses
            if (not self.args.options.noipv4
                and ("ipv4" in self.dns_env["dnservs"][servname].keys())):
                res["dnservs"][servname]["ipv4"] = {}

                # for every address
                for addr in self.dns_env["dnservs"][servname]["ipv4"].keys():
                    res["dnservs"][servname]["ipv4"][addr] = {}
                    # create a resolver
                    while resolver.pop_nameserver():
                        pass
                    resolver.push_nameserver(self.dns_env["dnservs"][servname]["ipv4"][addr]["rdf"])
                    if (resolver):
                        res["dnservs"][servname]["ipv4"][addr]["alive"] = True
                        pkt = resolver.query(domain, self.ldns.LDNS_RR_TYPE_SOA,
                            self.ldns.LDNS_RR_CLASS_IN, self.ldns.LDNS_RD)
                        if (not pkt):
                            self.err(0, "Resolver response error")
                            return False
                        res["dnservs"][servname]["ipv4"][addr]["aa"] = pkt.aa()
                        soa = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_SOA,
                            self.ldns.LDNS_SECTION_ANSWER)
                        if ((not soa) or (not soa.rr_count()) or (soa.rr(0).rd_count() != 7)):
                            self.err(0, "No SOA record, record empty or corrupt")
                            return False
                        res["dnservs"][servname]["ipv4"][addr]["SOA"] = dict(zip(soa_keys,[str(a) for a in soa.rr(0).rdfs()]))
                        self.verb(2, servname, "ipv4:", addr,
                            "name-server:", res["dnservs"][servname]["ipv4"][addr]["SOA"]["name-server"],
                            "email-addr:", res["dnservs"][servname]["ipv4"][addr]["SOA"]["email-addr"],
                            "sn",  res["dnservs"][servname]["ipv4"][addr]["SOA"]["sn"],
                            "refresh", res["dnservs"][servname]["ipv4"][addr]["SOA"]["refresh"],
                            "retry", res["dnservs"][servname]["ipv4"][addr]["SOA"]["retry"],
                            "expiry", res["dnservs"][servname]["ipv4"][addr]["SOA"]["expiry"],
                            "min", res["dnservs"][servname]["ipv4"][addr]["SOA"]["min"])
                    else:
                        res["dnservs"][servname]["ipv4"][addr]["alive"] = False

            # for ipv6 addresses
            if (self.args.options.resolver6
                and ("ipv6" in self.dns_env["dnservs"][servname].keys())):
                res["dnservs"][servname]["ipv6"] = {}
                
                # for every address
                for addr in self.dns_env["dnservs"][servname]["ipv6"].keys():
                    res["dnservs"][servname]["ipv6"][addr] = {}
                    # create a resolver
                    while resolver.pop_nameserver():
                        pass
                    resolver.push_nameserver(self.dns_env["dnservs"][servname]["ipv6"][addr]["rdf"])
                    if (resolver):
                        res["dnservs"][servname]["ipv6"][addr]["alive"] = True
                        pkt = resolver.query(domain, self.ldns.LDNS_RR_TYPE_SOA,
                            self.ldns.LDNS_RR_CLASS_IN, self.ldns.LDNS_RD)
                        if (not pkt):
                            self.err(0, "Resolver response error")
                            return False
                        res["dnservs"][servname]["ipv6"][addr]["aa"] = pkt.aa()
                        soa = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_SOA,
                            self.ldns.LDNS_SECTION_ANSWER)
                        if ((not soa) or (not soa.rr_count()) or (soa.rr(0).rd_count() != 7)):
                            self.err(0, "No SOA record, record empty or corrupt")
                            return False
                        res["dnservs"][servname]["ipv6"][addr]["SOA"] = dict(zip(soa_keys,[str(a) for a in soa.rr(0).rdfs()]))
                        self.verb(2, servname, "ipv6:", addr,
                            "name-server:", res["dnservs"][servname]["ipv6"][addr]["SOA"]["name-server"],
                            "email-addr:", res["dnservs"][servname]["ipv6"][addr]["SOA"]["email-addr"],
                            "sn",  res["dnservs"][servname]["ipv6"][addr]["SOA"]["sn"],
                            "refresh", res["dnservs"][servname]["ipv6"][addr]["SOA"]["refresh"],
                            "retry", res["dnservs"][servname]["ipv6"][addr]["SOA"]["retry"],
                            "expiry", res["dnservs"][servname]["ipv6"][addr]["SOA"]["expiry"],
                            "min", res["dnservs"][servname]["ipv6"][addr]["SOA"]["min"])
                    else:
                        res["dnservs"][servname]["ipv6"][addr]["alive"] = False

        self.verb(2, "")

        # second, check data in the tree structure various attributes

        # check whether all servers are responding
        all_alive = True
        for servname in res["dnservs"].keys():
            if ((not self.args.options.noipv4) and
                ("ipv4" in res["dnservs"][servname].keys())):
                for ipv4 in res["dnservs"][servname]["ipv4"].keys():
                    all_alive = all_alive and res["dnservs"][servname]["ipv4"][ipv4]["alive"]
                    if (res["dnservs"][servname]["ipv4"][ipv4]["alive"]):
                        self.verb(1, servname, ipv4, "responding")
                    else:
                        self.verb(1, servname, ipv4, "not responding")
            if (self.args.options.resolver6 and
                ("ipv6" in res["dnservs"][servname].keys())):
                for ipv6 in res["dnservs"][servname]["ipv6"].keys():
                    all_alive = all_alive and res["dnservs"][servname]["ipv6"][ipv6]["alive"]
                    if (res["dnservs"][servname]["ipv6"][ipv6]["alive"]):
                        self.verb(1, servname, ipv6, "responding")
                    else:
                        self.verb(1, servname, ipv6, "not responding")
        self.verb(1, "")

        if (all_alive):
            self.verb(0, "All DNS servers responding: yes")
        else:
            self.verb(0, "All DNS servers responding: no")
        self.verb(0, "")

        #check domain serial number
        all_same = True
        sn = None
        sn_list = []
        for servname in res["dnservs"].keys():
            if ((not self.args.options.noipv4) and
                ("ipv4") in res["dnservs"][servname].keys()):
                for ipv4 in res["dnservs"][servname]["ipv4"].keys():
                    if ("SOA" in res["dnservs"][servname]["ipv4"][ipv4].keys()):
                        serial = res["dnservs"][servname]["ipv4"][ipv4]["SOA"]["sn"]
                        self.verb(1, servname, ipv4, "sn:", serial)
                        if (not sn):
                            sn = serial
                            sn_list.append(serial)
                        else:
                            if (all_same):
                                if (sn != serial):
                                    all_same = False
                                    if (serial not in sn_list):
                                        sn_list.append(serial)
                            else:
                                if (serial not in sn_list):
                                    sn_list.append(serial)
            if (self.args.options.resolver6 and
                ("ipv6" in res["dnservs"][servname].keys())):
                for ipv6 in res["dnservs"][servname]["ipv6"].keys():
                    if ("SOA" in res["dnservs"][servname]["ipv6"][ipv6].keys()):
                        serial = res["dnservs"][servname]["ipv6"][ipv6]["SOA"]["sn"]
                        self.verb(1, servname, ipv6, "sn:", serial)
                        if (not sn):
                            sn = serial
                            sn_list.append(serial)
                        else:
                            if (all_same):
                                if (sn != serial):
                                    all_same = False
                                    if (serial not in sn_list):
                                        sn_list.append(serial)
                            else:
                                if (serial not in sn_list):
                                    sn_list.append(serial)
        self.verb(1, "")

        if (all_same):
            self.verb(0, "All DNS servers have the same serial number (", sn, "): yes")
        else:
            self.verb(0, "All DNS servers have the same serial number: no")
        self.verb(0, "")

        # check whether all servers return authoritative answers
        all_aa = True
        for servname in res["dnservs"].keys():
            if ((not self.args.options.noipv4) and
                ("ipv4") in res["dnservs"][servname].keys()):
                for ipv4 in res["dnservs"][servname]["ipv4"].keys():
                    all_aa = all_aa and res["dnservs"][servname]["ipv4"][ipv4]["aa"]
                    self.verb(1, servname, ipv4,
                        "authoritative:", res["dnservs"][servname]["ipv4"][ipv4]["aa"])
            if (self.args.options.resolver6 and
                ("ipv6" in res["dnservs"][servname].keys())):
                for ipv6 in res["dnservs"][servname]["ipv6"].keys():
                    all_aa = all_aa and res["dnservs"][servname]["ipv6"][ipv6]["aa"]
                    self.verb(1, servname, ipv6,
                        "authoritative:", res["dnservs"][servname]["ipv6"][ipv6]["aa"])
        self.verb(1, "")

        if (all_aa):
            self.verb(0, "All DNS servers are authoritative: yes")
        else:
            self.verb(0, "All DNS servers are authoritative: no")
        self.verb(0, "")

        # check domain mname
        all_same = True
        mn = None
        mn_list = []
        for servame in res["dnservs"].keys():
            if ((not self.args.options.noipv4) and
                ("ipv4" in res["dnservs"][servname].keys())):
                for ipv4 in res["dnservs"][servname]["ipv4"].keys():
                    if ("SOA" in res["dnservs"][servname]["ipv4"][ipv4].keys()):
                        mname = res["dnservs"][servname]["ipv4"][ipv4]["SOA"]["name-server"]
                        self.verb(1, servname, ipv4, "mname:", mname)
                        if (not mn):
                            mn = mname
                            mn_list.append(mname)
                        else:
                            if (all_same):
                                if (mn != mname):
                                    all_same = False
                                    if (mname not in mn_list):
                                        mn_list.append(mname)
                            else:
                                if (mname not in mn_list):
                                    mn_list.append(mname)
            if (self.args.options.resolver6 and
                ("ipv6" in res["dnservs"][servname].keys())):
                for ipv6 in res["dnservs"][servname]["ipv6"].keys():
                    if ("SOA" in res["dnservs"][servname]["ipv6"][ipv6].keys()):
                        mname = res["dnservs"][servname]["ipv6"][ipv6]["SOA"]["name-server"]
                        self.verb(1, servname, ipv6, "mname:", mname)
                        if (not mn):
                            mn = mname
                            mn_list.append(mname)
                        else:
                            if (all_same):
                                if (mn != mname):
                                    all_same = False
                                    if (mname not in mn_list):
                                        mn_list.append(mname)
                            else:
                                if (mname not in mn_list):
                                    mn_list.append(mname)
        self.verb(1, "")

        if (all_same):
            self.verb(0, "All DNS servers have the same primary server name (", mn, "): yes")
        else:
            self.verb(0, "All DNS servers have the same primary server name: no")
        self.verb(0, "")

        # check mname syntax
        all_ok = True
        for mn in mn_list:
            all_ok = all_ok and name_syntax(mn, self.dns_env["domain"])
            self.verb(1, "primary server name:", mn)
        if (all_ok):
            self.verb(0, "All primary server name values seem syntactically ok: yes")
        else:
            self.verb(0, "All primary server name values seem syntactically ok: no")
        self.verb(0, "")

        # check domain email
        all_same = True
        en = None
        en_list = []
        for servame in res["dnservs"].keys():
            if ((not self.args.options.noipv4) and
                ("ipv4" in res["dnservs"][servname].keys())):
                for ipv4 in res["dnservs"][servname]["ipv4"].keys():
                    if ("SOA" in res["dnservs"][servname]["ipv4"][ipv4].keys()):
                        ename = res["dnservs"][servname]["ipv4"][ipv4]["SOA"]["email-addr"]
                        self.verb(1, servname, ipv4, "email:", ename)
                        if (not en):
                            en = ename
                            en_list.append(ename)
                        else:
                            if (all_same):
                                if (en != ename):
                                    all_same = False
                                    if (ename not in en_list):
                                        en_list.append(ename)
                            else:
                                if (ename not in en_list):
                                    en_list.append(ename)
            if (self.args.options.resolver6 and
                ("ipv6" in res["dnservs"][servname].keys())):
                for ipv6 in res["dnservs"][servname]["ipv6"].keys():
                    if ("SOA" in res["dnservs"][servname]["ipv6"][ipv6].keys()):
                        ename = res["dnservs"][servname]["ipv6"][ipv6]["SOA"]["email-addr"]
                        self.verb(1, servname, ipv6, "email:", ename)
                        if (not en):
                            en = ename
                            en_list.append(ename)
                        else:
                            if (all_same):
                                if (en != ename):
                                    all_same = False
                                    if (ename not in en_list):
                                        en_list.append(ename)
                            else:
                                if (ename not in en_list):
                                    en_list.append(ename)
        self.verb(1, "")

        if (all_same):
            self.verb(0, "All DNS servers have the same email (", en, "): yes")
        else:
            self.verb(0, "All DNS servers have the same email: no")
        self.verb(0, "")

        # check email syntax
        all_ok = True
        no_at = True
        for en in en_list:
            all_ok = all_ok and name_syntax(en, self.dns_env["domain"])
            no_at = no_at and (en.find("@") < 0)
            self.verb(1, "email:", en)
        if (all_ok and no_at):
            self.verb(0, "All email values seem syntactically ok: yes")
        else:
            self.verb(0, "All email values seem syntactically ok: no")
        self.verb(0, "")

        self.set_my_modenv(res)
        return True

